Unit DBF1PLUS;
{****************************************************************}
{  dBase file read-write utility                 DATE: 11-NOV-94 }
{                                         Last change: 27-mar-95 }
{                                                                }
{  Denis A. Davydov                                              }
{****************************************************************}
INTERFACE
uses Crt,Objects;

const
 {consts for error message}
 cNoErrors   = 0;
 cIOError    = 1;
 cMemError   = 2;
 cRangeError = 3;
 cTypeError  = 4;
 cFormatError= 5;

Type
  {type of last date updating}
  TDate  = record
    YY : byte;
    MM : byte;
    DD : byte;
  end;
  {type of field descriptor}
  TField = record
    N : string[11]; {field name}
    T : char;       {field type (C,L,N,M,F)}
    L : byte;       {field length}
    D : byte;       {decimal places}
    Offset : word;  {offset field begin from record begin}
  end;
  PField = ^TField;
  {************************************************************
  Object TDBase automaticaly open *.dbf file and read header by
  DOS name.
  *************************************************************}
  TDBase = object
      FileName           : String;  {DOS file name}
      Version            : byte;    {version of dBase}
      LastDate           : TDate;   {date of last updating}
      RecNo              : longint; {count of records}
      RecLen             : word;    {size of record}
      FieldNo            : word;    {count of fields}
      errorcode          : word;    {error code value}
      constructor Init(fname: String);
      destructor  Done;
      procedure   GetField(N:word;var FF : TField);
      function    GetCell(Rec:longint;Field:word):string;
      procedure   error; virtual;    {error handler}
      procedure   Append; 	{append blank record to end of file}
      procedure   SetNCell(Rec:longint;Field:word;Val:Real); {write N cell}
      procedure   SetSCell(Rec:longint;Field:word;Val:String); {write S cell}
      procedure   Truncate(Rec:Longint); {delete records after Rec}
      procedure   Zap; {delete all records}
      procedure   Delete(Rec: Longint); {mark record as deleted}
      procedure   UnDelete(Rec: Longint); {clear delete flag}
      function    Deleted(Rec: Longint) : boolean;
      procedure   Date; {update last date change}
      procedure   CopyStructure(NewFileName : String); {copy ctructure to a new file}
    private
      function FieldAddr(N : word): PField;
    private
      procedure SetField(N:word;FF:TField);
    private
      procedure SetCell(Rec:longint;Field:word;Val:String); {write cell}
    private
      {deep internals}
      F                  : file; {file variable}
      _buffer            : array[1..10240] of char; {file buffer}
      HeaderSize         : word;         {size of header}
      Fa                 : Pointer;      {pointer to descriptor array}
  end; {TDBase}

IMPLEMENTATION

{************************************************************}
{                    TDBase.FieldAddr                        }
{                                                            }
{          Pointer function for internal use only            }
{************************************************************}
function TDBase.FieldAddr(N : word): PField;
begin
  if (N >= 1)and(N<=FieldNo) then
    FieldAddr := Ptr ( seg(Fa^), ofs(Fa^)+(N-1)*SizeOf(TField) )
  else
    begin
      FieldAddr := Nil;
      errorcode :=cRangeError;
      error;
    end;
end;

{************************************************************}
{                    TDbase.SetField                         }
{                                                            }
{Put a descriptor record to desriptor array internal use only}
{************************************************************}
procedure TDbase.SetField(N:word;FF:TField);
begin
   FieldAddr(N)^:=FF;
end;

{************************************************************}
{                     TDbase.GetField                        }
{                                                            }
{        Read a descriptor record from descriptor array      }
{************************************************************}
procedure TDbase.GetField(N:word;var FF: TField);
begin
   FF := FieldAddr(N)^;
end;

{************************************************************}
{                    TDbase.GetCell                          }
{                                                            }
{       Get a Cell ([Rec]:[Field]) from a dbase file         }
{************************************************************}
function TDbase.GetCell(Rec:longint;Field:word):string;
var
  fbuf : TField;
  sbuf : string;
  ii   : integer;
begin
  if (Rec > RecNo) or (Rec < 0) then
    begin
      errorcode := cRangeError;
      error;
    end;
  GetField(Field,fbuf);
  {$I-}
  seek(f,HeaderSize+(Rec-1)*RecLen+fbuf.offset);
  {$I+}
  if IOResult <> 0 then
    begin
      errorcode := cRangeError;
      error;
    end;
  sbuf :='';
  {}
  {$I-}
  BlockRead(f,_buffer,fbuf.L);
  {$I+}
  if IOResult <> 0 then
    begin
      errorcode := cIOError;
      error;
    end;

  for ii:=1 to fbuf.L do
  begin
    sbuf := sbuf + _buffer[ii];
    if _buffer[ii] = #0 then break;
  end;
  GetCell := sbuf;
end;
{************************************************************}
{                      TDBase.Date                           }
{                                                            }
{                       procedure                            }
{************************************************************}
procedure TDBase.Date;
begin
  RunError
end;

{************************************************************}
{                     TDBase.error                           }
{                                                            }
{                  Error handler VIRTUAL                     }
{************************************************************}
procedure TDBase.error;
begin
  if errorcode <> cNoErrors then
    begin
      TextAttr := $0F;
      ClrScr;
      writeln;
      Writeln(' - error in dBase object');
      Writeln('   file: ',FileName);
      Write('   error description: ');
      case errorcode of
        cIOError    : writeln('001: i/o error');
        cMemError   : writeln('002: memory allocation error');
        cRangeError : writeln('003: range error');
        cTypeError  : writeln('004: type mismatch');
        cFormatError: writeln('005: file format error');
      end; {case}
      HALT(10);
    end;
end;
{************************************************************}
{                      TDBase.Truncate                       }
{                                                            }
{                        procedure                           }
{************************************************************}
procedure TDBase.Truncate(Rec:Longint);
var
  RN : Longint;
  RNC : array[1..4] of char;
  RNB : array[1..4] of byte;
  i : byte;
begin
  {$I-}
  seek(F,HeaderSize+(Rec-1)*RecLen);
  {$I+}
  if IOResult <> 0 then
    begin
      errorcode := cIOError;
      error;
    end;
  System.truncate(F);
  RN := Rec;
     for i:=1 to 4 do
     begin
       RNB[i] := RN mod 256;
       RN := RN div 256;
     end;
     for i:=1 to 4 do
       RNC[i] := Chr(RNB[i]);
     Seek(F,4);   { <- area where is a record count}
     {$I-}
     BlockWrite(F,RNC,4);
     {$I+}
     if IOResult <> 0 then
       begin
         errorcode := cIOError;
         error;
       end;
end;

{************************************************************}
{                      TDBase.Zap                            }
{                                                            }
{                        procedure                           }
{************************************************************}
procedure TDBase.Zap;
begin
 TDBase.Truncate(1);
end;

{************************************************************}
{                        TDBase.Init                         }
{                                                            }
{                        Constructor                         }
{************************************************************}
constructor TDBase.Init(fname:String);
var
  Retcode : integer;
  cbuf    : array[0..31] of char;
  hbuf    : array[0..31] of byte;
  i,j     : integer;
  fbuf    : TField;
  offs    : word;
begin
  FA :=Nil;
  errorcode := cNoErrors;
  FileName := fname;
  Assign(F,FileName);
  {$I-}
  reset(F,1);
  {$I+}
  if IOREsult <> 0 then
    begin
      errorcode :=cIOError;
      error;
    end;
  {$I-}
  BlockRead(f,_buffer,32);
  {$I+}
  if IOREsult <> 0 then
    begin
      errorcode :=cIOError;
      error;
    end;
  for i := 1 to 31 do
    begin
      hbuf[i-1]:= Ord(_buffer[i]);
    end;
  Version     := hbuf[0];
  LastDate.YY := hbuf[1];
  LastDate.MM := hbuf[2];
  LastDate.DD := hbuf[3];
  HeaderSize  := hbuf[8]+hbuf[9]*256;
  RecLen      := hbuf[10]+hbuf[11]*256;
  RecNo       := hbuf[4]+hbuf[5]*256+hbuf[6]*256*256+hbuf[7]*256*256*256;
  FieldNo     := (hbuf[8]+hbuf[9]*256 -32 -1) div 32;
  {Header are readed now read descriptors}
  if MaxAvail < Sizeof(TField)*FieldNo then
    begin
      errorcode := cMemError;
      error;
    end;
  GetMem(Fa,Sizeof(TField)*FieldNo);
  offs := 1;
  for i := 1 to FieldNo do
    begin
      if (i*32) > FileSize(f) then
        begin
          errorcode := cRangeError;
          error;
        end;
      {$I-}
      seek(f,i*32);
      {$I+}
      if IOREsult <> 0 then
        begin
          errorcode :=cIOError;
          error;
        end;
      {$I-}
      BlockRead(F,_buffer,32);
      {$I+}
      if IOResult <> 0 then
        begin
          errorcode := cIOError;
          error;
        end;
      for j := 1 to 32 do
        begin
          hbuf[j-1] := Ord(_buffer[j]);
        end;
      fbuf.N := '';
      for j:=1 to 11 do
        begin
          if _buffer[j]= #0 then break;
          fbuf.N := fbuf.N+_buffer[j];
        end;
      fbuf.T := _buffer[12];
      fbuf.L := hbuf[16];
      fbuf.D := hbuf[17];
      fbuf.Offset := offs;
      offs := offs+fbuf.L;
      SetField(i,fbuf);
    end;
end; {constructor Init}

{************************************************************}
{                        TDBase.Done                         }
{                                                            }
{                        Destructor                          }
{************************************************************}
Destructor TDBase.Done;
begin
  if assigned(FA) then FreeMem(Fa,Sizeof(TField)*FieldNo);
  close(f);
end;

{************************************************************}
{                        TDBase.Append                       }
{                                                            }
{                        procedure                           }
{************************************************************}
Procedure TDbase.Append;
var
  i : word;
  RNB : array[1..4] of byte;
  RNC : array[1..4] of char;
  RN  : longint;
  cbuf: char;
begin
     Seek(F,FileSize(F));       {go to end of file}
     _buffer[1] := ' ';
     {$I-}
     BlockWrite(F,_buffer[1],1);
     {$I+}
     if IOResult <> 0 then
       begin
         errorcode := cIOError;
         error;
       end;
     for i:=1 to Reclen do _buffer[i] := #0;
     BlockWrite(F,_buffer,Reclen-1);
     if IOResult <> 0 then
       begin
         errorcode := cIOError;
         error;
       end;
     inc(RecNo);        {increment records count and write it}
(*     RecNo  := hbuf[4]+hbuf[5]*256+hbuf[6]*256*256+hbuf[7]*256*256*256;*)
     RN := RecNo;
     for i:=1 to 4 do
     begin
       RNB[i] := RN mod 256;
       RN := RN div 256;
     end;
     for i:=1 to 4 do
       RNC[i] := Chr(RNB[i]);
     Seek(F,4);   { <- area where is a record count}
     {$I-}
     BlockWrite(F,RNC,4);
     {$I+}
     if IOResult <> 0 then
       begin
         errorcode := cIOError;
         error;
       end;
end;
{************************************************************}
{                      TDBase.SetCell                        }
{                    write cell to a file                    }
{                  !!! for internal use !!!                  }
{                         procedure                          }
{************************************************************}
procedure TDBase.SetCell(Rec:longint;Field:word;Val:String);
var
  fbuf : TField;
  i    : byte;
  mLen : byte;
begin
  if (Rec > RecNo) or (Rec < 0) then
    begin
      errorcode := cRangeError;
      error;
    end;
  GetField(Field,fbuf);
  {$I-}
  seek(f,HeaderSize+(Rec-1)*RecLen+fbuf.offset);
  {$I+}
  if IOResult <> 0 then
    begin
      errorcode := cRangeError;
      error;
    end;
  {File positioned now writing}
  for i:=1 to fbuf.L do
    begin
      _buffer[i] :=Val[i];
      if Val[i] = #0 then break;
    end;
  {$I-}
  BlockWrite(F,_buffer,i);
  {$I+}
  if IOResult <> 0 then
    begin
      errorcode := cIOError;
      error;
    end;
    {Writted OK}
end;
{************************************************************}
{                     TDBase.SetNCell                        }
{      write numeric cell to a file (with type controll)     }
{                        procedure                           }
{************************************************************}
procedure TDBase.SetNCell(Rec:longint;Field:word;Val:Real);
var
  fbuf : TField;
  sbuf : string;
begin
  GetField(Field,fbuf);
  if fbuf.T <> 'N' then
    begin
      errorcode := cTypeError;
      error;
    end;
  Str(Val:fbuf.L:fbuf.D,sbuf);
  while sbuf[1]=' ' do
    System.delete(sbuf,1,1); {delete leading spases}
  if length(sbuf) < fbuf.L then sbuf:=sbuf+#0;
  SetCell(Rec,Field,sbuf);
end;
{************************************************************}
{                     TDBase.SetSCell                        }
{write string cell to a file (with type and length controll) }
{                        procedure                           }
{************************************************************}
procedure TDBase.SetSCell(Rec:longint;Field:word;Val:String);
var
  fbuf : TField;
  sbuf : string;
  i    : integer;
begin
  GetField(Field,fbuf);
  if fbuf.T <> 'C' then
    begin
      errorcode := cTypeError;
      error;
    end;
  if fbuf.L > Length(Val) then
    sbuf := Val + #0
  else
    sbuf := copy(Val,1,fbuf.L);
  SetCell(Rec,Field,sbuf);
end;
{************************************************************}
{                     TDBase.Delete                          }
{                                                            }
{                        procedure                           }
{************************************************************}
procedure TDBase.Delete(Rec:longint);
begin
  if (Rec > RecNo) or (Rec < 0) then
    begin
      errorcode := cRangeError;
      error;
    end;
  _buffer[1] := '*';
  {$I-}
  seek(f,HeaderSize+(Rec-1)*RecLen);
  BlockWrite(f,_buffer,1);
  {$I+}
  if IOResult <> 0 then
    begin
      errorcode := cIOError;
      error;
    end;
end;

{************************************************************}
{                     TDBase.UnDelete                        }
{                                                            }
{                        procedure                           }
{************************************************************}
procedure TDBase.UnDelete(Rec:longint);
begin
  if (Rec > RecNo) or (Rec < 0) then
    begin
      errorcode := cRangeError;
      error;
    end;
  _buffer[1] := ' ';
  {$I-}
  seek(f,HeaderSize+(Rec-1)*RecLen);
  BlockWrite(f,_buffer,1);
  {$I+}
  if IOResult <> 0 then
    begin
      errorcode := cIOError;
      error;
    end;
end;
{************************************************************}
{                     TDBase.CopyStructure                   }
{                                                            }
{                        procedure                           }
{************************************************************}
procedure TDBase.CopyStructure(NewFileName : String);
var
  Fnew : File of Char;
  Ch   : Char;
  I    : integer;
begin
  Assign(Fnew,NewFileName);
  {$I-}
  Rewrite(Fnew);
  {$I+}
  if IOResult <> 0 then
    begin
      errorcode := cIOError;
      error;
    end;
  seek(f,0);
  for I := 0 to HeaderSize do
    begin
      {$I-}
      BlockRead(f,ch,1);
      Write(fnew,ch);
      {$I+}
      if IOResult <> 0 then
        begin
          errorcode := cIOError;
          error;
        end;
    end;
  seek(fnew,4);
  Ch := #0;
  for i := 1 to 4 do
    begin
      {$I-}
      write(fnew,ch);
      {$I+}
      if IOResult <> 0 then
        begin
          errorcode := cIOError;
          error;
        end;
    end;
  Close(fnew);
end;
{************************************************************}
{                     TDBase.Deleted                         }
{            return true if Rec marked as Deleted            }
{                       function                             }
{************************************************************}
function TDBase.Deleted(Rec : Longint): boolean;
begin
  if (Rec > RecNo) or (Rec < 0) then
    begin
      errorcode := cRangeError;
      error;
    end;
  {$i-}
  seek(f,HeaderSize+(Rec-1)*RecLen);
  BlockRead(f,_buffer,1);
  {$i+}
  if IOResult <> 0 then
    begin
      errorcode := cIOError;
      error;
    end;
  Deleted := (_buffer[1] = '*');
end;

end.
